@@ -4,12 +4,16 @@
 */
 
 #include "motis2parser.h"
+
+#include "datatypes/rentalvehicleutil_p.h"
+#include "gbfs/gbfsvehicletypes.h"
 #include "geo/polylinedecoder_p.h"
 #include "gtfs/hvt.h"
 
 #include <KPublicTransport/Journey>
 #include <KPublicTransport/Location>
 #include <KPublicTransport/RentalVehicle>
+#include <KPublicTransport/StopInformation>
 #include <KPublicTransport/Stopover>
 
 #include <KCountry>
@@ -89,6 +93,9 @@ struct {
     { "FUNICULAR", Line::Mode::Funicular, JourneySection::PublicTransport, IndividualTransport::Walk, IndividualTransport::None, RentalVehicle::Unknown },
     { "AREAL_LIFT", Line::Mode::AerialLift, JourneySection::PublicTransport, IndividualTransport::Walk, IndividualTransport::None, RentalVehicle::Unknown },
     { "OTHER", Line::Mode::Unknown, JourneySection::PublicTransport, IndividualTransport::Walk, IndividualTransport::None, RentalVehicle::Unknown },
+    { "AERIAL_LIFT", Line::Mode::RapidTransit, JourneySection::PublicTransport, IndividualTransport::Walk, IndividualTransport::None, RentalVehicle::Unknown },
+    { "SUBURBAN", Line::Mode::RapidTransit, JourneySection::PublicTransport, IndividualTransport::Walk, IndividualTransport::None, RentalVehicle::Unknown },
+    { "RIDE_SHARING", Line::Mode::RideShare, JourneySection::PublicTransport, IndividualTransport::Car, IndividualTransport::None, RentalVehicle::Car },
 };
 
 struct {
@@ -192,6 +199,7 @@ Route Motis2Parser::parseRoute(const QJsonObject &obj) const
     }
 
     line.setOperatorName(obj.value("agencyName"_L1).toString());
+    line.setOperatorUrl(QUrl(obj.value("agencyUrl"_L1).toString()));
     line.setColor(QColor::fromString('#'_L1 + obj.value("routeColor"_L1).toString()));
     line.setTextColor(QColor::fromString('#'_L1 + obj.value("routeTextColor"_L1).toString()));
 
@@ -259,7 +267,12 @@ Journey Motis2Parser::parseItinerary(const QJsonObject &itinerary) const
         s.setMode((*it).jnyMode);
 
         const auto hasRealTime = leg.value("realTime"_L1).toBool();
-        const auto from = parsePlace(leg.value("from"_L1).toObject(), hasRealTime);
+        auto from = parsePlace(leg.value("from"_L1).toObject(), hasRealTime);
+        std::vector<Feature> features;
+        if (leg["mode"_L1] == "NIGHT_RAIL"_L1) {
+            features.push_back(Feature(Feature::SleeperCompartment));
+        }
+        from.setFeatures(std::move(features));
         s.setDeparture(from);
         const auto to = parsePlace(leg.value("to"_L1).toObject(), hasRealTime);
         s.setArrival(to);
@@ -301,6 +314,7 @@ Journey Motis2Parser::parseItinerary(const QJsonObject &itinerary) const
 
             RentalVehicleNetwork rvNetwork;
             rvNetwork.setName(rentalObj.value("systemName"_L1).toString());
+            rvNetwork.setUrl(QUrl(rentalObj.value("url"_L1).toString()));
 
             RentalVehicle rv;
             rv.setType((*it).rentalType);
@@ -484,6 +498,27 @@ std::vector<Location> Motis2Parser::parseLocations(const QByteArray &data) const
         l.setStreetAddress(QString(locObj.value("street"_L1).toString() + ' '_L1 + locObj.value("houseNumber"_L1).toString()).trimmed());
 
         parseTimeZone(locObj, l);
+
+        if (l.type() == Location::Stop) {
+            StopInformation stopInfo;
+            const auto modes = locObj.value("modes"_L1).toArray();
+            for (const auto &mode : modes) {
+                const auto it = std::ranges::find_if(mode_map, [&mode](const auto &m) {
+                    return QLatin1StringView(m.name) == mode;
+                });
+                if (it == std::end(mode_map)) {
+                    qWarning() << "Unknown mode:" << mode;
+                    continue;
+                }
+                Line l;
+                l.setMode((*it).lineMode);
+                stopInfo.addLine(l);
+            }
+            if (!stopInfo.lines().empty()) {
+                l.setData(stopInfo);
+            }
+        }
+
         result.push_back(std::move(l));
     }
 
@@ -507,6 +542,134 @@ std::vector<Location> Motis2Parser::parseMapStops(const QByteArray &data) const
     return result;
 }
 
+template <typename T>
+static void readRentalBookingUrls(const QJsonObject &obj, T &rental)
+{
+    rental.setWebBookingUrl(QUrl(obj.value("rentalUriWeb"_L1).toString()));
+#ifdef Q_OS_ANDROID
+    rental.setAppBookingUrl(QUrl(obj.value("rentalUriAndroid"_L1).toString()));
+#elif defined(Q_OS_IOS)
+    rental.setAppBookingUrl(QUrl(obj.value("rentalUriIOS"_L1).toString()));
+#endif
+}
+
+std::vector<Location> Motis2Parser::parseRentals(const QByteArray &data) const
+{
+    const auto rentalsObj = QJsonDocument::fromJson(data).object();
+
+    const auto providersArray = rentalsObj.value("providers"_L1).toArray();
+    std::unordered_map<QString, RentalVehicleNetwork> providers;
+    providers.reserve(providersArray.size());
+    std::unordered_map<QString, std::unordered_map<QString, GBFSVehicleType>> vehicleTypes;
+
+    for (const auto &providerV : providersArray) {
+        const auto providerObj = providerV.toObject();
+        RentalVehicleNetwork provider;
+        provider.setName(providerObj.value("name"_L1).toString());
+        if (const auto purchaseUrl = providerObj.value("purchaseUrl"_L1).toString(); !purchaseUrl.isEmpty()) {
+            provider.setUrl(QUrl(purchaseUrl));
+        } else {
+            provider.setUrl(QUrl(providerObj.value("url"_L1).toString()));
+        }
+        const auto providerId = providerObj.value("id"_L1).toString();
+
+        std::unordered_map<QString, GBFSVehicleType> providerVehicleTypes;
+        const auto typesArray = providerObj.value("vehicleTypes"_L1).toArray();
+        providerVehicleTypes.reserve(typesArray.size());
+        for (const auto &typeV : typesArray) {
+            const auto typeObj = typeV.toObject();
+
+            GBFSVehicleType vt;
+            vt.typeId = typeObj.value("id"_L1).toString();
+            vt.name = typeObj.value("name"_L1).toString();
+            vt.formFactor = GBFSVehicleType::parseFormFactor(typeObj.value("formFactor"_L1).toString());
+            vt.propulsionType = GBFSVehicleType::parsePropulsionType(typeObj.value("propulsionType"_L1).toString());
+
+            provider.setVehicleTypes(provider.vehicleTypes() | RentalVehicleUtil::fromGbfsVehicleType(vt));
+            providerVehicleTypes[vt.typeId] = std::move(vt);
+        }
+
+        providers[providerId] = std::move(provider);
+        vehicleTypes[providerId] = std::move(providerVehicleTypes);
+    }
+
+    const auto stations = rentalsObj.value("stations"_L1).toArray();
+    const auto vehicles = rentalsObj.value("vehicles"_L1).toArray();
+    std::vector<Location> result;
+    result.reserve(stations.size() + vehicles.size());
+
+    for (const auto &stationV : stations) {
+        const auto &stationObj = stationV.toObject();
+        const auto providerId = stationObj.value("providerId"_L1).toString();
+
+        Location l;
+        l.setType(Location::RentedVehicleStation);
+        l.setCoordinate(stationObj.value("lat"_L1).toDouble(), stationObj.value("lon"_L1).toDouble());
+        l.setName(stationObj.value("name"_L1).toString());
+        l.setIdentifier(m_locIdentifierType, providerId + ':'_L1 + stationObj.value("id"_L1).toString());
+        if (const auto addr = stationObj.value("address"_L1).toString(); !addr.isEmpty()) {
+            l.setStreetAddress(addr);
+        }
+
+        RentalVehicleStation s;
+        readRentalBookingUrls(stationObj, s);
+        s.setAvailableVehicles(stationObj.value("numVehiclesAvailable"_L1).toInt(-1));
+
+        s.setNetwork(providers[providerId]);
+        std::unordered_map<RentalVehicle::VehicleType, std::pair<int, int>> docks;
+        const auto availableVehicles = stationObj.value("vehicleTypesAvailable"_L1).toObject();
+        for (auto it = availableVehicles.begin(); it != availableVehicles.end(); ++it) {
+            const auto t = RentalVehicleUtil::fromGbfsVehicleType(vehicleTypes[providerId][it.key()]);
+            s.setAvailableVehicles(t, std::max(0, s.availableVehicles(t)) + it.value().toInt());
+            docks[t].first = s.availableVehicles(t);
+        }
+        const auto availableDocks = stationObj.value("vehicleDocksAvailable"_L1).toObject();
+        for (auto it = availableDocks.begin(); it != availableDocks.end(); ++it) {
+            const auto t = RentalVehicleUtil::fromGbfsVehicleType(vehicleTypes[providerId][it.key()]);
+            docks[t].second += it.value().toInt();
+        }
+        if (!availableDocks.isEmpty()) {
+            for (auto it = docks.begin(); it != docks.end(); ++it) {
+                s.setCapacity((*it).first, (*it).second.first + (*it).second.second);
+            }
+        }
+
+        l.setData(std::move(s));
+        result.push_back(std::move(l));
+    }
+
+    for (const auto &vehicleV : vehicles) {
+        const auto vehicleObj = vehicleV.toObject();
+        const auto providerId = vehicleObj.value("providerId"_L1).toString();
+
+        if (vehicleObj.value("isReserved"_L1).toBool() || vehicleObj.value("isDisabled"_L1).toBool() || !vehicleObj.value("stationId"_L1).toString().isEmpty()) {
+            continue;
+        }
+
+        Location l;
+        l.setType(Location::RentedVehicle);
+        l.setCoordinate(vehicleObj.value("lat"_L1).toDouble(), vehicleObj.value("lon"_L1).toDouble());
+        l.setIdentifier(m_locIdentifierType, providerId + ':'_L1 + vehicleObj.value("id"_L1).toString());
+
+        RentalVehicle v;
+        readRentalBookingUrls(vehicleObj, v);
+        v.setNetwork(providers[providerId]);
+
+        const auto &vt = vehicleTypes[providerId][vehicleObj.value("typeId"_L1).toString()];
+        v.setType(RentalVehicleUtil::fromGbfsVehicleType(vt));
+        if (!vt.name.isEmpty()) {
+            l.setName(vt.name);
+        } else {
+            l.setName(v.network().name());
+        }
+        l.setData(std::move(v));
+
+        result.push_back(std::move(l));
+    }
+
+    return result;
+}
+
 void Motis2Parser::parseCursors(const QJsonObject &obj)
 {
     m_previousPageCursor = obj.value("previousPageCursor"_L1).toString();
